@@ -21,164 +21,24 @@
 #include "kcompletion_p.h"
 
 #include <QDebug>
-#include <QtCore/QMutableVectorIterator>
 #include <QCollator>
 
-class KCompletionPrivate
+void KCompletionPrivate::init()
 {
-public:
-    KCompletionPrivate()
-        : myCompletionMode(KCompletion::CompletionPopup)
-        , myTreeRoot(new KCompTreeNode)
-        , myBeep(true)
-        , myIgnoreCase(false)
-        , myHasMultipleMatches(false)
-        , myRotationIndex(0)
-    {
-    }
-    ~KCompletionPrivate()
-    {
-        delete myTreeRoot;
-    }
-    // list used for nextMatch() and previousMatch()
-    KCompletionMatchesWrapper matches;
-
-    KCompletion::CompletionMode myCompletionMode;
-
-    KCompletion::CompOrder myOrder;
-    QString                myLastString;
-    QString                myLastMatch;
-    QString                myCurrentMatch;
-    KCompTreeNode         *myTreeRoot;
-    //QStringList            myRotations;
-    bool                   myBeep : 1;
-    bool                   myIgnoreCase : 1;
-    bool                   myHasMultipleMatches;
-    int                    myRotationIndex;
-};
-
-KCompletion::KCompletion()
-    : d(new KCompletionPrivate)
-{
-    setOrder(Insertion);
+    completionMode = KCompletion::CompletionPopup;
+    treeNodeAllocator = KCompTreeNode::allocator(); // keep strong-ref to allocator instance
+    treeRoot = new KCompTreeNode;
+    beep = true;
+    ignoreCase = false;
+    hasMultipleMatches = false;
+    rotationIndex = 0;
 }
 
-KCompletion::~KCompletion()
+void KCompletionPrivate::addWeightedItem(const QString &item)
 {
-    delete d;
-}
-
-void KCompletion::setOrder(CompOrder order)
-{
-    d->myOrder = order;
-    d->matches.setSorting(order);
-}
-
-KCompletion::CompOrder KCompletion::order() const
-{
-    return d->myOrder;
-}
-
-void KCompletion::setIgnoreCase(bool ignoreCase)
-{
-    d->myIgnoreCase = ignoreCase;
-}
-
-bool KCompletion::ignoreCase() const
-{
-    return d->myIgnoreCase;
-}
-
-void KCompletion::setItems(const QStringList &items)
-{
-    clear();
-    insertItems(items);
-}
-
-void KCompletion::insertItems(const QStringList &items)
-{
-    bool weighted = (d->myOrder == Weighted);
-    QStringList::ConstIterator it;
-    if (weighted) {   // determine weight
-        for (it = items.begin(); it != items.end(); ++it) {
-            addWeightedItem(*it);
-        }
-    } else {
-        for (it = items.begin(); it != items.end(); ++it) {
-            addItem(*it, 0);
-        }
-    }
-}
-
-QStringList KCompletion::items() const
-{
-    KCompletionMatchesWrapper list; // unsorted
-    bool addWeight = (d->myOrder == Weighted);
-    extractStringsFromNode(d->myTreeRoot, QString(), &list, addWeight);
-
-    return list.list();
-}
-
-bool KCompletion::isEmpty() const
-{
-    return (d->myTreeRoot->childrenCount() == 0);
-}
-
-void KCompletion::postProcessMatch(QString *) const
-{
-}
-
-void KCompletion::postProcessMatches(QStringList *) const
-{
-}
-
-void KCompletion::postProcessMatches(KCompletionMatches *) const
-{
-}
-
-void KCompletion::addItem(const QString &item)
-{
-    d->matches.clear();
-    d->myRotationIndex = 0;
-    d->myLastString.clear();
-
-    addItem(item, 0);
-}
-
-void KCompletion::addItem(const QString &item, uint weight)
-{
-    if (item.isEmpty()) {
-        return;
-    }
-
-    KCompTreeNode *node = d->myTreeRoot;
-    uint len = item.length();
-
-    bool sorted = (d->myOrder == Sorted);
-    bool weighted = ((d->myOrder == Weighted) && weight > 1);
-
-    // knowing the weight of an item, we simply add this weight to all of its
-    // nodes.
-
-    for (uint i = 0; i < len; i++) {
-        node = node->insert(item.at(i), sorted);
-        if (weighted) {
-            node->confirm(weight - 1);    // node->insert() sets weighting to 1
-        }
-    }
-
-    // add 0x0-item as delimiter with evtl. weight
-    node = node->insert(0x0, true);
-    if (weighted) {
-        node->confirm(weight - 1);
-    }
-//     qDebug("*** added: %s (%i)", item.toLatin1().constData(), node->weight());
-}
-
-void KCompletion::addWeightedItem(const QString &item)
-{
-    if (d->myOrder != Weighted) {
-        addItem(item, 0);
+    Q_Q(KCompletion);
+    if (order != KCompletion::Weighted) {
+        q->addItem(item, 0);
         return;
     }
 
@@ -197,268 +57,16 @@ void KCompletion::addWeightedItem(const QString &item)
         len = index; // only insert until the ':'
     }
 
-    addItem(item.left(len), weight);
+    q->addItem(item.left(len), weight);
     return;
 }
 
-void KCompletion::removeItem(const QString &item)
-{
-    d->matches.clear();
-    d->myRotationIndex = 0;
-    d->myLastString.clear();
-
-    d->myTreeRoot->remove(item);
-}
-
-void KCompletion::clear()
-{
-    d->matches.clear();
-    d->myRotationIndex = 0;
-    d->myLastString.clear();
-
-    delete d->myTreeRoot;
-    d->myTreeRoot = new KCompTreeNode;
-}
-
-QString KCompletion::makeCompletion(const QString &string)
-{
-    if (d->myCompletionMode == CompletionNone) {
-        return QString();
-    }
-
-    //qDebug() << "KCompletion: completing: " << string;
-
-    d->matches.clear();
-    d->myRotationIndex = 0;
-    d->myHasMultipleMatches = false;
-    d->myLastMatch = d->myCurrentMatch;
-
-    // in Shell-completion-mode, emit all matches when we get the same
-    // complete-string twice
-    if (d->myCompletionMode == CompletionShell &&
-            string == d->myLastString) {
-        // Don't use d->matches since calling postProcessMatches()
-        // on d->matches here would interfere with call to
-        // postProcessMatch() during rotation
-
-        findAllCompletions(string, &d->matches, d->myHasMultipleMatches);
-        QStringList l = d->matches.list();
-        postProcessMatches(&l);
-        emit matches(l);
-
-        return QString();
-    }
-
-    QString completion;
-    // in case-insensitive popup mode, we search all completions at once
-    if (d->myCompletionMode == CompletionPopup ||
-            d->myCompletionMode == CompletionPopupAuto) {
-        findAllCompletions(string, &d->matches, d->myHasMultipleMatches);
-        if (!d->matches.isEmpty()) {
-            completion = d->matches.first();
-        }
-    } else {
-        completion = findCompletion(string);
-    }
-
-    if (d->myHasMultipleMatches) {
-        emit multipleMatches();
-    }
-
-    d->myLastString = string;
-    d->myCurrentMatch = completion;
-
-    postProcessMatch(&completion);
-
-    if (!string.isEmpty()) {   // only emit match when string is not empty
-        //qDebug() << "KCompletion: Match: " << completion;
-        emit match(completion);
-    }
-
-    return completion;
-}
-
-QStringList KCompletion::substringCompletion(const QString &string) const
-{
-    // get all items in the tree, eventually in sorted order
-    KCompletionMatchesWrapper allItems(d->myOrder);
-    extractStringsFromNode(d->myTreeRoot, QString(), &allItems, false);
-
-    QStringList list = allItems.list();
-
-    // subStringMatches is invoked manually, via a shortcut
-    if (list.isEmpty()) {
-        return list;
-    }
-
-    if (string.isEmpty()) {   // shortcut
-        postProcessMatches(&list);
-        return list;
-    }
-
-    QStringList matches;
-    QStringList::ConstIterator it = list.constBegin();
-
-    for (; it != list.constEnd(); ++it) {
-        QString item = *it;
-        if (item.indexOf(string, 0, Qt::CaseInsensitive) != -1) {     // always case insensitive
-            postProcessMatch(&item);
-            matches.append(item);
-        }
-    }
-
-    return matches;
-}
-
-void KCompletion::setCompletionMode(CompletionMode mode)
-{
-    d->myCompletionMode = mode;
-}
-
-KCompletion::CompletionMode KCompletion::completionMode() const
-{
-    return d->myCompletionMode;
-}
-
-QStringList KCompletion::allMatches()
-{
-    // Don't use d->matches since calling postProcessMatches()
-    // on d->matches here would interfere with call to
-    // postProcessMatch() during rotation
-    KCompletionMatchesWrapper matches(d->myOrder);
-    bool dummy;
-    findAllCompletions(d->myLastString, &matches, dummy);
-    QStringList l = matches.list();
-    postProcessMatches(&l);
-    return l;
-}
-
-KCompletionMatches KCompletion::allWeightedMatches()
-{
-    // Don't use d->matches since calling postProcessMatches()
-    // on d->matches here would interfere with call to
-    // postProcessMatch() during rotation
-    KCompletionMatchesWrapper matches(d->myOrder);
-    bool dummy;
-    findAllCompletions(d->myLastString, &matches, dummy);
-    KCompletionMatches ret(matches);
-    postProcessMatches(&ret);
-    return ret;
-}
-
-QStringList KCompletion::allMatches(const QString &string)
-{
-    KCompletionMatchesWrapper matches(d->myOrder);
-    bool dummy;
-    findAllCompletions(string, &matches, dummy);
-    QStringList l = matches.list();
-    postProcessMatches(&l);
-    return l;
-}
-
-KCompletionMatches KCompletion::allWeightedMatches(const QString &string)
-{
-    KCompletionMatchesWrapper matches(d->myOrder);
-    bool dummy;
-    findAllCompletions(string, &matches, dummy);
-    KCompletionMatches ret(matches);
-    postProcessMatches(&ret);
-    return ret;
-}
-
-void KCompletion::setSoundsEnabled(bool enable)
-{
-    d->myBeep = enable;
-}
-
-bool KCompletion::soundsEnabled() const
-{
-    return d->myBeep;
-}
-
-bool KCompletion::hasMultipleMatches() const
-{
-    return d->myHasMultipleMatches;
-}
-
-/////////////////////////////////////////////////////
-///////////////// tree operations ///////////////////
-
-QString KCompletion::nextMatch()
-{
-    QString completion;
-    d->myLastMatch = d->myCurrentMatch;
-
-    if (d->matches.isEmpty()) {
-        findAllCompletions(d->myLastString, &d->matches, d->myHasMultipleMatches);
-        if (!d->matches.isEmpty()) {
-            completion = d->matches.first();
-        }
-        d->myCurrentMatch = completion;
-        d->myRotationIndex = 0;
-        postProcessMatch(&completion);
-        emit match(completion);
-        return completion;
-    }
-
-    QStringList matches = d->matches.list();
-    d->myLastMatch = matches[ d->myRotationIndex++ ];
-
-    if (d->myRotationIndex == matches.count()) {
-        d->myRotationIndex = 0;
-    }
-
-    completion = matches[ d->myRotationIndex ];
-    d->myCurrentMatch = completion;
-    postProcessMatch(&completion);
-    emit match(completion);
-    return completion;
-}
-
-const QString &KCompletion::lastMatch() const
-{
-    return d->myLastMatch;
-}
-
-QString KCompletion::previousMatch()
-{
-    QString completion;
-    d->myLastMatch = d->myCurrentMatch;
-
-    if (d->matches.isEmpty()) {
-        findAllCompletions(d->myLastString, &d->matches, d->myHasMultipleMatches);
-        if (!d->matches.isEmpty()) {
-            completion = d->matches.last();
-        }
-        d->myCurrentMatch = completion;
-        d->myRotationIndex = 0;
-        postProcessMatch(&completion);
-        emit match(completion);
-        return completion;
-    }
-
-    QStringList matches = d->matches.list();
-    d->myLastMatch = matches[ d->myRotationIndex ];
-
-    if (d->myRotationIndex == 0) {
-        d->myRotationIndex = matches.count();
-    }
-
-    d->myRotationIndex--;
-
-    completion = matches[ d->myRotationIndex ];
-    d->myCurrentMatch = completion;
-    postProcessMatch(&completion);
-    emit match(completion);
-    return completion;
-}
-
 // tries to complete "string" from the tree-root
-QString KCompletion::findCompletion(const QString &string)
+QString KCompletionPrivate::findCompletion(const QString &string)
 {
     QChar ch;
     QString completion;
-    const KCompTreeNode *node = d->myTreeRoot;
+    const KCompTreeNode *node = treeRoot;
 
     // start at the tree-root and try to find the search-string
     for (int i = 0; i < string.length(); i++) {
@@ -485,11 +93,11 @@ QString KCompletion::findCompletion(const QString &string)
     // if multiple matches and auto-completion mode
     // -> find the first complete match
     if (node && node->childrenCount() > 1) {
-        d->myHasMultipleMatches = true;
+        hasMultipleMatches = true;
 
-        if (d->myCompletionMode == CompletionAuto) {
-            d->myRotationIndex = 1;
-            if (d->myOrder != Weighted) {
+        if (completionMode == KCompletion::CompletionAuto) {
+            rotationIndex = 1;
+            if (order != KCompletion::Weighted) {
                 while ((node = node->firstChild())) {
                     if (!node->isNull()) {
                         completion += *node;
@@ -529,7 +137,7 @@ QString KCompletion::findCompletion(const QString &string)
     return completion;
 }
 
-void KCompletion::findAllCompletions(const QString &string,
+void KCompletionPrivate::findAllCompletions(const QString &string,
                                      KCompletionMatchesWrapper *matches,
                                      bool &hasMultipleMatches) const
 {
@@ -539,15 +147,15 @@ void KCompletion::findAllCompletions(const QString &string,
         return;
     }
 
-    if (d->myIgnoreCase) {   // case insensitive completion
-        extractStringsFromNodeCI(d->myTreeRoot, QString(), string, matches);
+    if (ignoreCase) {   // case insensitive completion
+        extractStringsFromNodeCI(treeRoot, QString(), string, matches);
         hasMultipleMatches = (matches->count() > 1);
         return;
     }
 
     QChar ch;
     QString completion;
-    const KCompTreeNode *node = d->myTreeRoot;
+    const KCompTreeNode *node = treeRoot;
 
     // start at the tree-root and try to find the search-string
     for (int i = 0; i < string.length(); i++) {
@@ -586,7 +194,7 @@ void KCompletion::findAllCompletions(const QString &string,
     }
 }
 
-void KCompletion::extractStringsFromNode(const KCompTreeNode *node,
+void KCompletionPrivate::extractStringsFromNode(const KCompTreeNode *node,
         const QString &beginning,
         KCompletionMatchesWrapper *matches,
         bool addWeight) const
@@ -601,7 +209,7 @@ void KCompletion::extractStringsFromNode(const KCompTreeNode *node,
     QString w;
 
     // loop thru all children
-    for (KCompTreeNode *cur = list->begin(); cur; cur = cur->next) {
+    for (KCompTreeNode *cur = list->begin(); cur; cur = cur->m_next) {
         string = beginning;
         node = cur;
         if (!node->isNull()) {
@@ -633,7 +241,7 @@ void KCompletion::extractStringsFromNode(const KCompTreeNode *node,
     }
 }
 
-void KCompletion::extractStringsFromNodeCI(const KCompTreeNode *node,
+void KCompletionPrivate::extractStringsFromNodeCI(const KCompTreeNode *node,
         const QString &beginning,
         const QString &restString,
         KCompletionMatchesWrapper *matches) const
@@ -668,6 +276,402 @@ void KCompletion::extractStringsFromNodeCI(const KCompTreeNode *node,
     }
 }
 
+KCompletion::KCompletion()
+    : d_ptr(new KCompletionPrivate(this))
+{
+    Q_D(KCompletion);
+    d->init();
+    setOrder(Insertion);
+}
+
+KCompletion::~KCompletion()
+{
+}
+
+void KCompletion::setOrder(CompOrder order)
+{
+    Q_D(KCompletion);
+    d->order = order;
+    d->matches.setSorting(order);
+}
+
+KCompletion::CompOrder KCompletion::order() const
+{
+    Q_D(const KCompletion);
+    return d->order;
+}
+
+void KCompletion::setIgnoreCase(bool ignoreCase)
+{
+    Q_D(KCompletion);
+    d->ignoreCase = ignoreCase;
+}
+
+bool KCompletion::ignoreCase() const
+{
+    Q_D(const KCompletion);
+    return d->ignoreCase;
+}
+
+void KCompletion::setItems(const QStringList &itemList)
+{
+    clear();
+    insertItems(itemList);
+}
+
+void KCompletion::insertItems(const QStringList &items)
+{
+    Q_D(KCompletion);
+    bool weighted = (d->order == Weighted);
+    QStringList::ConstIterator it;
+    if (weighted) {   // determine weight
+        for (it = items.begin(); it != items.end(); ++it) {
+            d->addWeightedItem(*it);
+        }
+    } else {
+        for (it = items.begin(); it != items.end(); ++it) {
+            addItem(*it, 0);
+        }
+    }
+}
+
+QStringList KCompletion::items() const
+{
+    Q_D(const KCompletion);
+    KCompletionMatchesWrapper list; // unsorted
+    bool addWeight = (d->order == Weighted);
+    d->extractStringsFromNode(d->treeRoot, QString(), &list, addWeight);
+
+    return list.list();
+}
+
+bool KCompletion::isEmpty() const
+{
+    Q_D(const KCompletion);
+    return (d->treeRoot->childrenCount() == 0);
+}
+
+void KCompletion::postProcessMatch(QString *) const
+{
+}
+
+void KCompletion::postProcessMatches(QStringList *) const
+{
+}
+
+void KCompletion::postProcessMatches(KCompletionMatches *) const
+{
+}
+
+void KCompletion::addItem(const QString &item)
+{
+    Q_D(KCompletion);
+    d->matches.clear();
+    d->rotationIndex = 0;
+    d->lastString.clear();
+
+    addItem(item, 0);
+}
+
+void KCompletion::addItem(const QString &item, uint weight)
+{
+    Q_D(KCompletion);
+    if (item.isEmpty()) {
+        return;
+    }
+
+    KCompTreeNode *node = d->treeRoot;
+    uint len = item.length();
+
+    bool sorted = (d->order == Sorted);
+    bool weighted = ((d->order == Weighted) && weight > 1);
+
+    // knowing the weight of an item, we simply add this weight to all of its
+    // nodes.
+
+    for (uint i = 0; i < len; i++) {
+        node = node->insert(item.at(i), sorted);
+        if (weighted) {
+            node->confirm(weight - 1);    // node->insert() sets weighting to 1
+        }
+    }
+
+    // add 0x0-item as delimiter with evtl. weight
+    node = node->insert(0x0, true);
+    if (weighted) {
+        node->confirm(weight - 1);
+    }
+//     qDebug("*** added: %s (%i)", item.toLatin1().constData(), node->weight());
+}
+
+void KCompletion::removeItem(const QString &item)
+{
+    Q_D(KCompletion);
+    d->matches.clear();
+    d->rotationIndex = 0;
+    d->lastString.clear();
+
+    d->treeRoot->remove(item);
+}
+
+void KCompletion::clear()
+{
+    Q_D(KCompletion);
+    d->matches.clear();
+    d->rotationIndex = 0;
+    d->lastString.clear();
+
+    delete d->treeRoot;
+    d->treeRoot = new KCompTreeNode;
+}
+
+QString KCompletion::makeCompletion(const QString &string)
+{
+    Q_D(KCompletion);
+    if (d->completionMode == CompletionNone) {
+        return QString();
+    }
+
+    //qDebug() << "KCompletion: completing: " << string;
+
+    d->matches.clear();
+    d->rotationIndex = 0;
+    d->hasMultipleMatches = false;
+    d->lastMatch = d->currentMatch;
+
+    // in Shell-completion-mode, emit all matches when we get the same
+    // complete-string twice
+    if (d->completionMode == CompletionShell &&
+            string == d->lastString) {
+        // Don't use d->matches since calling postProcessMatches()
+        // on d->matches here would interfere with call to
+        // postProcessMatch() during rotation
+
+        d->findAllCompletions(string, &d->matches, d->hasMultipleMatches);
+        QStringList l = d->matches.list();
+        postProcessMatches(&l);
+        emit matches(l);
+
+        return QString();
+    }
+
+    QString completion;
+    // in case-insensitive popup mode, we search all completions at once
+    if (d->completionMode == CompletionPopup ||
+            d->completionMode == CompletionPopupAuto) {
+        d->findAllCompletions(string, &d->matches, d->hasMultipleMatches);
+        if (!d->matches.isEmpty()) {
+            completion = d->matches.first();
+        }
+    } else {
+        completion = d->findCompletion(string);
+    }
+
+    if (d->hasMultipleMatches) {
+        emit multipleMatches();
+    }
+
+    d->lastString = string;
+    d->currentMatch = completion;
+
+    postProcessMatch(&completion);
+
+    if (!string.isEmpty()) {   // only emit match when string is not empty
+        //qDebug() << "KCompletion: Match: " << completion;
+        emit match(completion);
+    }
+
+    return completion;
+}
+
+QStringList KCompletion::substringCompletion(const QString &string) const
+{
+    Q_D(const KCompletion);
+    // get all items in the tree, eventually in sorted order
+    KCompletionMatchesWrapper allItems(d->order);
+    d->extractStringsFromNode(d->treeRoot, QString(), &allItems, false);
+
+    QStringList list = allItems.list();
+
+    // subStringMatches is invoked manually, via a shortcut
+    if (list.isEmpty()) {
+        return list;
+    }
+
+    if (string.isEmpty()) {   // shortcut
+        postProcessMatches(&list);
+        return list;
+    }
+
+    QStringList matches;
+    QStringList::ConstIterator it = list.constBegin();
+
+    for (; it != list.constEnd(); ++it) {
+        QString item = *it;
+        if (item.indexOf(string, 0, Qt::CaseInsensitive) != -1) {     // always case insensitive
+            postProcessMatch(&item);
+            matches.append(item);
+        }
+    }
+
+    return matches;
+}
+
+void KCompletion::setCompletionMode(CompletionMode mode)
+{
+    Q_D(KCompletion);
+    d->completionMode = mode;
+}
+
+KCompletion::CompletionMode KCompletion::completionMode() const
+{
+    Q_D(const KCompletion);
+    return d->completionMode;
+}
+
+QStringList KCompletion::allMatches()
+{
+    Q_D(KCompletion);
+    // Don't use d->matches since calling postProcessMatches()
+    // on d->matches here would interfere with call to
+    // postProcessMatch() during rotation
+    KCompletionMatchesWrapper matches(d->order);
+    bool dummy;
+    d->findAllCompletions(d->lastString, &matches, dummy);
+    QStringList l = matches.list();
+    postProcessMatches(&l);
+    return l;
+}
+
+KCompletionMatches KCompletion::allWeightedMatches()
+{
+    Q_D(KCompletion);
+    // Don't use d->matches since calling postProcessMatches()
+    // on d->matches here would interfere with call to
+    // postProcessMatch() during rotation
+    KCompletionMatchesWrapper matches(d->order);
+    bool dummy;
+    d->findAllCompletions(d->lastString, &matches, dummy);
+    KCompletionMatches ret(matches);
+    postProcessMatches(&ret);
+    return ret;
+}
+
+QStringList KCompletion::allMatches(const QString &string)
+{
+    Q_D(KCompletion);
+    KCompletionMatchesWrapper matches(d->order);
+    bool dummy;
+    d->findAllCompletions(string, &matches, dummy);
+    QStringList l = matches.list();
+    postProcessMatches(&l);
+    return l;
+}
+
+KCompletionMatches KCompletion::allWeightedMatches(const QString &string)
+{
+    Q_D(KCompletion);
+    KCompletionMatchesWrapper matches(d->order);
+    bool dummy;
+    d->findAllCompletions(string, &matches, dummy);
+    KCompletionMatches ret(matches);
+    postProcessMatches(&ret);
+    return ret;
+}
+
+void KCompletion::setSoundsEnabled(bool enable)
+{
+    Q_D(KCompletion);
+    d->beep = enable;
+}
+
+bool KCompletion::soundsEnabled() const
+{
+    Q_D(const KCompletion);
+    return d->beep;
+}
+
+bool KCompletion::hasMultipleMatches() const
+{
+    Q_D(const KCompletion);
+    return d->hasMultipleMatches;
+}
+
+/////////////////////////////////////////////////////
+///////////////// tree operations ///////////////////
+
+QString KCompletion::nextMatch()
+{
+    Q_D(KCompletion);
+    QString completion;
+    d->lastMatch = d->currentMatch;
+
+    if (d->matches.isEmpty()) {
+        d->findAllCompletions(d->lastString, &d->matches, d->hasMultipleMatches);
+        if (!d->matches.isEmpty()) {
+            completion = d->matches.first();
+        }
+        d->currentMatch = completion;
+        d->rotationIndex = 0;
+        postProcessMatch(&completion);
+        emit match(completion);
+        return completion;
+    }
+
+    QStringList matches = d->matches.list();
+    d->lastMatch = matches[ d->rotationIndex++ ];
+
+    if (d->rotationIndex == matches.count()) {
+        d->rotationIndex = 0;
+    }
+
+    completion = matches[ d->rotationIndex ];
+    d->currentMatch = completion;
+    postProcessMatch(&completion);
+    emit match(completion);
+    return completion;
+}
+
+const QString &KCompletion::lastMatch() const
+{
+    Q_D(const KCompletion);
+    return d->lastMatch;
+}
+
+QString KCompletion::previousMatch()
+{
+    Q_D(KCompletion);
+    QString completion;
+    d->lastMatch = d->currentMatch;
+
+    if (d->matches.isEmpty()) {
+        d->findAllCompletions(d->lastString, &d->matches, d->hasMultipleMatches);
+        if (!d->matches.isEmpty()) {
+            completion = d->matches.last();
+        }
+        d->currentMatch = completion;
+        d->rotationIndex = 0;
+        postProcessMatch(&completion);
+        emit match(completion);
+        return completion;
+    }
+
+    QStringList matches = d->matches.list();
+    d->lastMatch = matches[ d->rotationIndex ];
+
+    if (d->rotationIndex == 0) {
+        d->rotationIndex = matches.count();
+    }
+
+    d->rotationIndex--;
+
+    completion = matches[ d->rotationIndex ];
+    d->currentMatch = completion;
+    postProcessMatch(&completion);
+    emit match(completion);
+    return completion;
+}
+
 /////////////////////////////////
 /////////
 
@@ -679,10 +683,10 @@ void KCompletion::extractStringsFromNodeCI(const KCompTreeNode *node,
 KCompTreeNode::~KCompTreeNode()
 {
     // delete all children
-    KCompTreeNode *cur = myChildren.begin();
+    KCompTreeNode *cur = m_children.begin();
     while (cur) {
-        KCompTreeNode *next = cur->next;
-        delete myChildren.remove(cur);
+        KCompTreeNode *next = cur->m_next;
+        delete m_children.remove(cur);
         cur = next;
     }
 }
@@ -698,24 +702,24 @@ KCompTreeNode *KCompTreeNode::insert(const QChar &ch, bool sorted)
         // FIXME, first (slow) sorted insertion implementation
         if (sorted) {
             KCompTreeNode *prev = 0;
-            KCompTreeNode *cur = myChildren.begin();
+            KCompTreeNode *cur = m_children.begin();
             while (cur) {
                 if (ch > *cur) {
                     prev = cur;
-                    cur = cur->next;
+                    cur = cur->m_next;
                 } else {
                     break;
                 }
             }
             if (prev) {
-                myChildren.insert(prev, child);
+                m_children.insert(prev, child);
             } else {
-                myChildren.prepend(child);
+                m_children.prepend(child);
             }
         }
 
         else {
-            myChildren.append(child);
+            m_children.append(child);
         }
     }
 
@@ -754,145 +758,59 @@ void KCompTreeNode::remove(const QString &str)
     for (; i >= 1; i--) {
         parent = deletables.at(i - 1);
         child = deletables.at(i);
-        if (child->myChildren.count() == 0) {
-            delete parent->myChildren.remove(child);
+        if (child->m_children.count() == 0) {
+            delete parent->m_children.remove(child);
         }
     }
 }
 
 QStringList KCompletionMatchesWrapper::list() const
 {
-    if (sortedList && dirty) {
-        sortedList->sort();
-        dirty = false;
+    if (m_sortedList && m_dirty) {
+        m_sortedList->sort();
+        m_dirty = false;
 
-        stringList.clear();
+        m_stringList.clear();
 
         // high weight == sorted last -> reverse the sorting here
         QList<KSortableItem<QString> >::const_iterator it;
-        for (it = sortedList->constBegin(); it != sortedList->constEnd(); ++it) {
-            stringList.prepend((*it).value());
+        for (it = m_sortedList->constBegin(); it != m_sortedList->constEnd(); ++it) {
+            m_stringList.prepend((*it).value());
         }
-    } else if (compOrder == KCompletion::Sorted) {
+    } else if (m_compOrder == KCompletion::Sorted) {
         QCollator c;
         c.setCaseSensitivity(Qt::CaseSensitive);
-        qStableSort(stringList.begin(), stringList.end(), c);
+        qStableSort(m_stringList.begin(), m_stringList.end(), c);
     }
 
-    return stringList;
-}
-
-class KCompletionMatchesPrivate
-{
-public:
-    KCompletionMatchesPrivate(bool sort)
-        : sorting(sort)
-    {}
-    bool sorting;
-};
-
-KCompletionMatches::KCompletionMatches(const KCompletionMatches &o)
-    : KSortableList<QString, int>(),
-      d(new KCompletionMatchesPrivate(o.d->sorting))
-{
-    *this = KCompletionMatches::operator=(o);
-}
-
-KCompletionMatches &KCompletionMatches::operator=(const KCompletionMatches &o)
-{
-    if (*this == o) {
-        return *this;
-    }
-    KCompletionMatchesList::operator=(o);
-    d->sorting = o.d->sorting;
-
-    return *this;
-}
-
-KCompletionMatches::KCompletionMatches(bool sort_P)
-    : d(new KCompletionMatchesPrivate(sort_P))
-{
-}
-
-KCompletionMatches::KCompletionMatches(const KCompletionMatchesWrapper &matches)
-    : d(new KCompletionMatchesPrivate(matches.sorting()))
-{
-    if (matches.sortedList != 0L) {
-        KCompletionMatchesList::operator=(*matches.sortedList);
-    } else {
-        const QStringList l = matches.list();
-        for (QStringList::ConstIterator it = l.begin();
-                it != l.end();
-                ++it) {
-            prepend(KSortableItem<QString, int>(1, *it));
-        }
-    }
-}
-
-KCompletionMatches::~KCompletionMatches()
-{
-    delete d;
-}
-
-QStringList KCompletionMatches::list(bool sort_P) const
-{
-    if (d->sorting && sort_P) {
-        const_cast< KCompletionMatches * >(this)->sort();
-    }
-    QStringList stringList;
-    // high weight == sorted last -> reverse the sorting here
-    for (ConstIterator it = begin(); it != end(); ++it) {
-        stringList.prepend((*it).value());
-    }
-    return stringList;
-}
-
-bool KCompletionMatches::sorting() const
-{
-    return d->sorting;
-}
-
-void KCompletionMatches::removeDuplicates()
-{
-    Iterator it1, it2;
-    for (it1 = begin(); it1 != end(); ++it1) {
-        for ((it2 = it1), ++it2; it2 != end();) {
-            if ((*it1).value() == (*it2).value()) {
-                // use the max height
-                (*it1).first = qMax((*it1).key(), (*it2).key());
-                it2 = erase(it2);
-                continue;
-            }
-            ++it2;
-        }
-    }
+    return m_stringList;
 }
 
 void KCompTreeNodeList::append(KCompTreeNode *item)
 {
     m_count++;
-    if (!last) {
-        last = item;
-        last->next = 0;
-        first = item;
+    if (!m_last) {
+        m_last = item;
+        m_last->m_next = 0;
+        m_first = item;
         return;
     }
-    last->next = item;
-    item->next = 0;
-    last = item;
+    m_last->m_next = item;
+    item->m_next = 0;
+    m_last = item;
 }
 
 void KCompTreeNodeList::prepend(KCompTreeNode *item)
 {
     m_count++;
-    if (!last) {
-        last = item;
-        last->next = 0;
-        first = item;
+    if (!m_last) {
+        m_last = item;
+        m_last->m_next = 0;
+        m_first = item;
         return;
     }
-    item->next = first;
-    first = item;
+    item->m_next = m_first;
+    m_first = item;
 }
 
 void KCompTreeNodeList::insert(KCompTreeNode *after, KCompTreeNode *item)
@@ -904,35 +822,35 @@ void KCompTreeNodeList::insert(KCompTreeNode *after, KCompTreeNode *item)
 
     m_count++;
 
-    item->next = after->next;
-    after->next = item;
+    item->m_next = after->m_next;
+    after->m_next = item;
 
-    if (after == last) {
-        last = item;
+    if (after == m_last) {
+        m_last = item;
     }
 }
 
 KCompTreeNode *KCompTreeNodeList::remove(KCompTreeNode *item)
 {
-    if (!first || !item) {
+    if (!m_first || !item) {
         return 0;
     }
     KCompTreeNode *cur = 0;
 
-    if (item == first) {
-        first = first->next;
+    if (item == m_first) {
+        m_first = m_first->m_next;
     } else {
-        cur = first;
-        while (cur && cur->next != item) {
-            cur = cur->next;
+        cur = m_first;
+        while (cur && cur->m_next != item) {
+            cur = cur->m_next;
         }
         if (!cur) {
             return 0;
         }
-        cur->next = item->next;
+        cur->m_next = item->m_next;
     }
-    if (item == last) {
-        last = cur;
+    if (item == m_last) {
+        m_last = cur;
     }
     m_count--;
     return item;
@@ -940,12 +858,12 @@ KCompTreeNode *KCompTreeNodeList::remove(KCompTreeNode *item)
 
 KCompTreeNode *KCompTreeNodeList::at(uint index) const
 {
-    KCompTreeNode *cur = first;
+    KCompTreeNode *cur = m_first;
     while (index-- && cur) {
-        cur = cur->next;
+        cur = cur->m_next;
     }
     return cur;
 }
 
-KZoneAllocator KCompTreeNode::alloc(8192);
+QSharedPointer<KZoneAllocator> KCompTreeNode::m_alloc(new KZoneAllocator(8*1024));
 
